# pyright: basic
from typing import (
    Any, Sequence, MutableSequence, List, Mapping, MutableMapping, Dict, Deque,
    Set, Tuple, DefaultDict, OrderedDict as OrderedDict_, FrozenSet, Type, Iterable, cast
)
from collections import abc, deque, defaultdict, OrderedDict, Counter, ChainMap
from datetime import date, time, datetime
from decimal import Decimal
from types import NoneType
from pytest import raises as assert_raises, fixture

from typingutils import TypeParameter, UnionParameter, AnyType, get_type_name, isinstance_typing, is_type
from tests.other_impl.isinstance import comparison_generator

instance_assertions: list[tuple[object | tuple[object], tuple[TypeParameter | UnionParameter, ...], bool]] = [
    (type, (
        object,
        Type[Any],
        type[Any]
        ), False),
    (None, (
        object,
        NoneType,
        Type[Any],
        type[Any]), True),
    ("abcdefg", (
        object,
        Type[Any],
        type[Any],
        str,
        abc.Collection,
        abc.Collection[Any],
        abc.Sequence,
        abc.Sequence[Any],
        Sequence,
        Sequence[Any]), True),
    (123, (
        object,
        Type[Any],
        type[Any],
        int), True),
    (10.12345, (
        object,
        Type[Any],
        type[Any],
        float), True),
    (Decimal(10.12345), (
        object,
        Type[Any],
        type[Any],
        Decimal), True),
    ((True, False), (
        object,
        Type[Any],
        type[Any],
        bool,
        int), True),
    (date(2000, 1, 1), (
        object,
        Type[Any],
        type[Any],
        date), True),
    (time(10, 0, 0), (
        object,
        Type[Any],
        type[Any],
        time), True),
    (datetime(2000, 1, 1, 10, 0, 0), (
        object,
        Type[Any],
        type[Any],
        datetime,
        date), True),
    (([], ["a", "b", "c"]), (
        object,
        Type[Any],
        type[Any],
        list,
        list[Any],
        List,
        List[Any],
        abc.Collection,
        abc.Collection[Any],
        abc.Sequence,
        abc.Sequence[Any],
        abc.MutableSequence,
        abc.MutableSequence[Any],
        Sequence,
        Sequence[Any],
        MutableSequence,
        MutableSequence[Any]), True),
    (deque((1,2,3), maxlen=10), (
        object,
        Type[Any],
        type[Any],
        deque,
        deque[Any],
        Deque,
        Deque[Any],
        abc.Collection,
        abc.Collection[Any],
        abc.Sequence,
        abc.Sequence[Any],
        abc.MutableSequence,
        abc.MutableSequence[Any],
        Sequence,
        Sequence[Any],
        MutableSequence,
        MutableSequence[Any]), True),
    (({}, { "a": 1}), (
        object,
        Type[Any],
        type[Any],
        dict,
        dict[Any, Any],
        Dict,
        Dict[Any, Any],
        abc.Mapping,
        abc.Mapping[Any, Any],
        abc.MutableMapping,
        abc.MutableMapping[Any, Any],
        abc.Collection,
        abc.Collection[Any],
        Mapping,
        Mapping[Any, Any],
        MutableMapping,
        MutableMapping[Any, Any]), True),
    (defaultdict(list), (
        object,
        Type[Any],
        type[Any],
        dict,
        dict[Any, Any],
        defaultdict,
        defaultdict[Any, Any],
        Dict,
        Dict[Any, Any],
        DefaultDict,
        DefaultDict[Any, Any],
        abc.Mapping,
        abc.Mapping[Any, Any],
        abc.MutableMapping,
        abc.MutableMapping[Any, Any],
        abc.Collection,
        abc.Collection[Any],
        Mapping,
        Mapping[Any, Any],
        MutableMapping,
        MutableMapping[Any, Any]), True),
    (OrderedDict(), (
        object,
        Type[Any],
        type[Any],
        dict,
        dict[Any, Any],
        Dict,
        Dict[Any, Any],
        OrderedDict,
        OrderedDict[Any, Any],
        OrderedDict_,
        OrderedDict_[Any, Any],
        abc.Mapping,
        abc.Mapping[Any, Any],
        abc.MutableMapping,
        abc.MutableMapping[Any, Any],
        abc.Collection,
        abc.Collection[Any],
        Mapping,
        Mapping[Any, Any],
        MutableMapping,
        MutableMapping[Any, Any]), True),
    (ChainMap(), (
        object,
        Type[Any],
        type[Any],
        ChainMap,
        ChainMap[Any, Any],
        abc.Mapping,
        abc.Mapping[Any, Any],
        abc.MutableMapping,
        abc.MutableMapping[Any, Any],
        abc.Collection,
        abc.Collection[Any],
        Mapping,
        Mapping[Any, Any],
        MutableMapping,
        MutableMapping[Any, Any]), True),
    ({ 1, 2, 3}, (
        object,
        Type[Any],
        type[Any],
        set,
        set[Any],
        Set,
        Set[Any],
        abc.Set,
        abc.Set[Any],
        abc.Collection,
        abc.Collection[Any]), True),
    (frozenset((1,2,3)), (
        object,
        Type[Any],
        type[Any],
        abc.Set,
        abc.Set[Any],
        abc.Collection,
        abc.Collection[Any],
        FrozenSet,
        FrozenSet[Any]), True),
    ((), (
        object,
        Type[Any],
        type[Any],
        tuple,
        Tuple,
        Tuple[Any],
        Sequence,
        Sequence[Any]), True)
]

all_types = set( type_ for _, types, _ in instance_assertions for type_ in types )
all_instances = tuple( (instance, is_inst) for instances, _, is_inst in instance_assertions for instance in cast(Iterable[object], instances if isinstance(instances, tuple) else (instances,)) )


@fixture(scope = "module")
def comparisons():
    impl: dict[str, list[tuple[str, str]]] = defaultdict(lambda: [])
    yield impl

    print("\n")

    for key in impl:
        count = 0
        failed = 0
        for comparison, result in impl[key]:
            if "Error" in result:
                failed += 1
                print(f"{comparison} ==> FAILED")
            else:
                print(f"{comparison} ==> {result}")
                count +=1

        print(f"Comparison with {key}: {count} differences and {failed} failed\n")


def test_all_types(comparisons: dict[str, list[tuple[str, str]]]):
    for type_ in all_types:
        result = is_type(type_)

        if type_ is object:
            assert not result
        else:
            assert result

        result = isinstance_typing(type_, type_)

        if type_ is object:
            assert result # object is always an instance of object
        else:
            assert not result

        for impl, result_comparison in comparison_generator(type_, type_):
            if result != result_comparison:
                comparisons[impl].append((f"Comparing {impl}.isinstance({type_}, {get_type_name(type_)})", f"{result_comparison} != {result}"))


def test_all_instances(comparisons: dict[str, list[tuple[str, str]]]):
    assert not isinstance_typing(list)
    assert isinstance_typing(object)
    assert not isinstance_typing(type)
    assert isinstance_typing("abc")
    assert isinstance_typing([1,2,3])

    for instance, is_inst in all_instances:
        result = isinstance_typing(instance)

        if is_inst:
            print(f"Testing isinstance_typing({instance}) ==> {result}")
            assert result
        else:
            print(f"Testing !isinstance_typing({instance}) ==> {result}")
            assert not result



def test_explicit_assertions(comparisons: dict[str, list[tuple[str, str]]]):
    for instances, types, is_inst in instance_assertions:
        negations = all_types.difference(types)

        for instance in cast(Iterable[object], instances if isinstance(instances, tuple) else (instances,)):
            for type_ in types:
                result = isinstance_typing(instance, type_)

                print(f"Testing isinstance_typing({instance}, {get_type_name(type_)}) ==> {result}")
                assert result is not None
                assert result == True

                for impl, result_comparison in comparison_generator(instance, type_):
                    if result != result_comparison:
                        comparisons[impl].append((f"Comparing {impl}.isinstance({instance}, {get_type_name(type_)})", f"{result_comparison} != {result}"))


            for type_ in negations:
                result = isinstance_typing(instance, type_)
                print(f"Testing !isinstance_typing({instance}, {get_type_name(type_)}) ==> {result}")

                assert result is not None
                assert result == False

                for impl, result_comparison in comparison_generator(instance, type_):
                    if result != result_comparison:
                        comparisons[impl].append((f"Comparing {impl}.isinstance({instance}, {get_type_name(type_)})", f"{result_comparison} = {result}"))


def test_multiple(comparisons: dict[str, list[tuple[str, str]]]):
    for obj, cls, expected in cast(tuple[tuple[object, Tuple[TypeParameter|UnionParameter, ...], bool]], (
        ("abc", (str, int, bool), True),
        ("abc", (float, int, bool), False),
        ("abc", (float, int, bool), False),
        ("abc", (float, str|int, bool), True),
    )):

        result = isinstance_typing(obj, cls)
        print(f"Testing isinstance_typing({obj}, {cls}) ==> {result}")
        assert result == expected

        for impl, result_comparison in comparison_generator(obj, cls):
            if result != result_comparison:
                comparisons[impl].append((f"Comparing {impl}.isinstance({obj}, {get_type_name(cast(AnyType, cls))})", f"{result_comparison} != {result}"))

def test_builtin_generic_types(comparisons: dict[str, list[tuple[str, str]]]):
    for obj, cls, expected in cast(tuple[tuple[object, Tuple[TypeParameter|UnionParameter, ...], bool]], (
        ([], list[Any], True),
        ([], list, True),
        ([], list[str], False),
        (["a", "b", "c"], list[Any], True),
        (["a", "b", "c"], list, True),
        (["a", "b", "c"], list[str], False),
        ({}, dict[Any, Any], True),
        ({}, dict, True),
        ({}, dict[str, Any], False),
        ({"a": 1, "b": 2}, dict[Any, Any], True),
        ({"a": 1, "b": 2}, dict, True),
        ({"a": 1, "b": 2}, dict[str, Any], False),
    )):

        result = isinstance_typing(obj, cls)
        print(f"Testing isinstance_typing({obj}, {cls}) ==> {result}")
        assert result == expected

        for impl, result_comparison in comparison_generator(obj, cls):
            if result != result_comparison:
                comparisons[impl].append((f"Comparing {impl}.isinstance({obj}, {get_type_name(cast(AnyType, cls))})", f"{result_comparison} != {result}"))


def test_recursive(comparisons: dict[str, list[tuple[str, str]]]):
    class IterableInt(abc.Iterable):
        def __iter__(self):
            return iter((1, 2, 3))
    class IterableStr(abc.Iterable):
        def __iter__(self):
            return iter(("a", "b", "c"))

    for obj, cls, expected in cast(tuple[tuple[object, Tuple[TypeParameter|UnionParameter, ...], bool]], (
        ("abc", Iterable[str], True),
        (123, Iterable[str], False),
        ((), tuple[Any], True),
        ((), tuple, True),
        ((), tuple[str], True),
        (("a", "b", "c"), tuple[Any], True),
        (("a", "b", "c"), tuple, True),
        (("a", "b", "c"), tuple[str], True),
        (("a", "b", "c"), tuple[int], False),
        ([], list[Any], True),
        ([], list, True),
        ([], list[str], True),
        (["a", "b", "c"], list[Any], True),
        (["a", "b", "c"], list, True),
        (["a", "b", "c"], list[str], True),
        (["a", "b", "c"], list[int], False),
        ({}, dict[Any, Any], True),
        ({}, dict, True),
        ({}, dict[str, Any], True),
        ({"a": 1, "b": 2}, dict[Any, Any], True),
        ({"a": 1, "b": 2}, dict, True),
        ({"a": 1, "b": 2}, dict[str, Any], True),
        ({"a": 1, "b": 2}, dict[str, int], True),
        ({"a": 1, "b": 2}, dict[str, str], False),
        ({"a": 1, "b": 2}, dict[int, int], False),
        ({"a", "b"}, set[Any], True),
        ({"a", "b"}, set[str], True),
        ({"a", "b"}, set[int], False),
        (IterableInt(), Iterable, True),
        (IterableInt(), Iterable[int], True),
        (IterableInt(), Iterable[str], False),
        (IterableStr(), Iterable, True),
        (IterableStr(), Iterable[str], True),
        (IterableStr(), Iterable[int], False),
    )):

        result = isinstance_typing(obj, cls, recursive=True)
        print(f"Testing isinstance_typing({obj}, {cls}, recursive=True) ==> {result}")
        if result != expected:
            isinstance_typing(obj, cls, recursive=True)
        assert result == expected

        for impl, result_comparison in comparison_generator(obj, cls):
            if result != result_comparison:
                comparisons[impl].append((f"Comparing {impl}.isinstance({obj}, {get_type_name(cast(AnyType, cls))})", f"{result_comparison} != {result}"))

        with assert_raises(Exception):
            isinstance_typing(iter(IterableInt()), Iterable[int], recursive=True)


from typing import Callable, Type, TypeVar, Any, Generic, Union, NamedTuple, overload, cast
from types import UnionType, NoneType
from collections import abc
from inspect import stack as get_stack

from typingutils.core.attributes import  ORIGIN, ORIGINAL_CLASS, ARGS, TYPE_PARAMS
from typingutils.core.types import (
    TypeParameter, TypeVarParameter, UnionParameter, AnyType, SetOfAny, TypeArgs,
    is_subscripted_generic_type, is_generic_type, get_generic_origin, issubclass_typing
)

class TypeCheck(NamedTuple):
    is_type: bool
    is_generic_type: bool
    is_subscripted_generic_type: bool

def get_original_class(obj: Any) -> TypeParameter:
    """
    Returns the original generic type from a class instance.
    This is useful for generic types because instances of these doesn't derive from them,
    thus having no generic arguments specified. Will even work when called from within a constructor of a class.

    Notes:
        Won't work with builtin generic types like list and tuple.

    Examples:
        class GenClass[T]:
            def __init__(self): # self == GenClass
                self.org = get_original_class(self) # org == GenClass[x]

        t = GenClass[str]() # type(t) == GenClass
        g = get_original_class(t) # g == GenClass[str]

    Args:
        obj (Any): An instance of a class.

    Returns:
        type: The objects original class if any - otherwise the class itself is returned.
    """

    cls: TypeParameter = type(obj) if not is_type(obj) else obj # pyright: ignore[reportUnknownVariableType]

    if hasattr(obj, ORIGINAL_CLASS):
        return getattr(obj, ORIGINAL_CLASS)
    elif is_generic_type(cls) or is_subscripted_generic_type(cls):
        stack = get_stack()
        for frame in stack[2:]:
            if frame.filename == __file__:
                continue # pragma: no cover
            locals = frame[0].f_locals
            if "self" in locals:
                self = locals["self"]
                if hasattr(self, ORIGIN) and getattr(self, ORIGIN) == cls:
                    return self

    return cls

def _extract_args(obj: Any) -> tuple[tuple[TypeParameter, ...] | None, tuple[TypeVarParameter, ...] | None, tuple[TypeParameter | UnionParameter, ...] | None]:
    """
    Extracts arguments from a generic object or type.

    Examples:
        T = TypeVar('T', bound=str)
        class GenClass(Generic[T]): pass
        params, args, types = _extract_args(GenClass) # => (~T<str>, None, (str,))
        params, args, types = _extract_args(GenClass[str]) # => (None, (str,), (str,))

    Args:
        obj (Any): A class or an instance of a class.

    Returns:
        tuple[
            tuple[TypeParameter, ...] | None,
            tuple[TypeVarParameter, ...] | None,
            tuple[TypeParameter | UnionParameter, ...] | None
        ]: Three sequences corresponding to parameters, arguments and types.
    """

    for attr in (ARGS, TYPE_PARAMS):
        if hasattr(obj, attr):
            from typingutils.core.types import get_types_from_typevar

            args = tuple(
                (
                    arg,
                    get_types_from_typevar(arg) if isinstance(arg, TypeVar) else arg,
                    isinstance(arg, TypeVar)
                )
                for arg in cast(tuple[Any], getattr(obj, attr))
            )
            parameters = tuple( arg for arg, _, typevar in args if typevar )
            arguments = tuple( arg for arg, _, typevar in args if not typevar )
            parameter_types = tuple( arg if not typevar else value for arg, value, typevar in args if not typevar )

            # in python 3.13 certain types may contain both typevars and types in the __args__ attribyte,
            # case in point typing.ContextManager[T] which has ` ~T, bool |None ` except of the expected ` ~T `
            # which is why either parameters or arguments must be None when returned

            if parameters and any(parameters):
                return parameters, None, parameter_types
            elif arguments and any(arguments):
                return None, arguments, parameter_types
            else:
                return None, None, None # pragma: no cover

    return None, None, None

def get_generic_arguments(obj: Any) -> TypeArgs:
    """
    Returns the type arguments used to create a subscripted generic type.
    Will even work when called from within a constructor of the class.

    Notes:
        The class must inherit Generic[T], and it must me the first inherited type to work.

    Examples:
        T = TypeVar('T')
        class GenClass(Generic[T]): pass
        a = get_generic_arguments(GenClass[str]) => (str,)

    Args:
        obj (Any): A type, object or instance of an object.

    Returns:
        TypeArgs: A sequence of types.
    """

    _, args, _ = _extract_args(obj)
    if args is not None:
        return args
    elif not is_type(obj):
        orig_class = get_original_class(obj)
        if orig_class is not type(obj):
            return cast(Callable[[Any], tuple[type, ...]], get_generic_arguments)(orig_class)
    elif isinstance(obj, type) and Generic in obj.__bases__:
        orig_class = get_original_class(obj)
        if orig_class != obj:
            return cast(Callable[[Any], tuple[type, ...]], get_generic_arguments)(orig_class)


    return ()


def is_type(obj: Any) -> bool:
    """
    Checks if object is a type (or a generic type) or not.

    Notes:
        TypeVar's aren't recognized as types.

    Args:
        obj (Any): A type, object or instance of an object.

    Returns:
        bool: A boolean value indicating if object is a type.
    """

    if type(obj) is TypeVar:
        return False
    elif obj is object:
        return False
    elif obj is Any:
        return False
    elif isinstance(obj, UnionType) or get_generic_origin(obj) == Union:
        return True

    return isinstance(obj, type) or is_generic_type(obj) or is_subscripted_generic_type(obj)



def check_type(obj: Any) -> TypeCheck:
    """
    Checks if object is a type.

    Args:
        obj (Any): An instance of an object or a type.

    Returns:
        TypeCheck: Returns a TypeCheck tuple.
    """
    if is_subscripted_generic_type(obj):
        return TypeCheck(True, False, True)
    elif is_generic_type(obj):
        return TypeCheck(True, True, False)
    elif is_type(obj):
        return TypeCheck(True, False, False)
    else:
        return TypeCheck(False, False, False)


@overload
def isinstance_typing(obj: Any) -> bool:
    """
    Checks if object is an instance of an object or not.

    Args:
        obj (Any): A type, object or instance of an object.

    Returns:
        bool: A boolean value indicating if object is an instance of an object or not.
    """
    ...
@overload
def isinstance_typing(obj: Any, cls: AnyType | TypeArgs, *, recursive: bool = False) -> bool:
    """
    Checks if object is an instance of the specified type/types. This implementation
    works similarly to the builtin isinstance(), but supports generics as well.

    Args:
        obj (Any): An object or instance of an object.
        cls (type): A type or sequence of types.

    Returns:
        bool: A boolean value indicating if object is an instance of the specified type/types.
    """
    ...
def isinstance_typing(obj: Any, cls: AnyType | TypeArgs | None = None, *, recursive: bool = False) -> bool:
    obj_is_type = is_type(obj)

    if cls is None and not obj_is_type:
        return True
    elif obj is cls is object:
        return True # object is always an instance of itself
    elif cls is Any:
        return True # all objects are an instance of Any
    elif obj is cls:
        return False # an object is never an instance of itself (unless it is object - see previous line)
    elif obj is type and cls in (object, type[Any], Type, Type[Any]):
        return True # types are only derived from object and type[Any]
    elif type(obj) is type and cls in (object, type, type[Any], Type, Type[Any]):
        return True # all classes are derived from type and object
    elif type(obj) is not type and cls in (object, type[Any], Type[Any]):
        return True # all class instances are derived from type and object
    elif type(obj) is NoneType and cls in (object, type[Any], Type[Any], NoneType):
        return True # None is derived from type and object

    if isinstance(cls, tuple):
        for cls1 in cls:
            if isinstance_typing(obj, cls1):
                return True
        return False

    if obj_is_type:
        return cls is object # all other types are only an instance of object
    if not obj_is_type and cls is type:
        return False # only other types are derived from type object
    if get_generic_origin(cast(type, cls)) is Union:
        for cls1 in getattr(cls, ARGS):
            if isinstance_typing(obj, cls1):
                return True
        return False
    if issubclass_typing(type(obj), cast(type, cls)): # pyright: ignore[reportUnknownArgumentType]
        return True

    obj_has_orig_cls_attr = hasattr(obj, ORIGINAL_CLASS)
    cls_has_args_attr = hasattr(cls, ARGS)
    cls_has_origin_attr = hasattr(cls, ORIGIN)

    if obj_has_orig_cls_attr and getattr(obj, ORIGINAL_CLASS) != type:
        origin = getattr(obj, ORIGINAL_CLASS)
        if origin == cls:
            return True

    if not obj_has_orig_cls_attr and cls_has_origin_attr and cls_has_args_attr and set(get_generic_arguments(cls)) == SetOfAny:
        cls = getattr(cls, ORIGIN)

    if not is_subscripted_generic_type(cast(TypeParameter | UnionParameter, cls)) and not is_generic_type(cast(TypeParameter, type(obj))):
        return isinstance(obj, cast(type, cls))

    elif isinstance(cls, TypeVar):
        from typingutils.core.types import get_types_from_typevar
        return isinstance_typing(obj, get_types_from_typevar(cls))


    if recursive and ( args := get_generic_arguments(cls) ):
        cls_origin = cast(type[Any], getattr(cls, ORIGIN) if cls_has_origin_attr else cls)

        if issubclass(cls_origin, abc.Sequence) and isinstance(obj, abc.Sequence) and isinstance_typing(obj, cast(type[Any], cls_origin)):
            for item in cast(abc.Sequence[Any], obj):
                if not isinstance_typing(item, args[0]):
                    return False
            return True

        if issubclass(cls_origin, abc.Mapping) and isinstance(obj, abc.Mapping) and isinstance_typing(obj, cast(type[Any], cls_origin)):
            for key in cast(abc.Mapping[Any, Any], obj):
                if not isinstance_typing(key, args[0]):
                    return False
                if not isinstance_typing(obj[key], args[1]):
                    return False
            return True

        if issubclass(cls_origin, abc.Set) and isinstance(obj, abc.Set) and isinstance_typing(obj, cast(type[Any], cls_origin)):
            for item in cast(abc.Set[Any], obj):
                if not isinstance_typing(item, args[0]):
                    return False
            return True


        if issubclass(cls_origin, abc.Iterable) and isinstance(obj, abc.Iterable) and isinstance_typing(obj, cast(type[Any], cls_origin)):
            if isinstance_typing(obj, abc.Iterator):
                raise Exception("Recursive instance checks on iterators cannot be done at it would deplete the iterator for further use")
            for item in cast(abc.Iterable[Any], obj):
                if not isinstance_typing(item, args[0]):
                    return False
            return True

    return False


# pyright: basic
from typing import Any, TypeVar, cast
from pytest import raises as assert_raises, fixture
import sys

from typingutils import get_generic_parameters, is_generic_function, AnyFunction, TypeArgs
from typingutils.internal import get_executing_function
from typingutils.core.attributes import TYPE_PARAMS

def test_is_generic_function():

    def generic_function(arg: int) -> int:
        ...

    assert not is_generic_function(generic_function)

    # generic functions are available in python 3.12
    setattr(generic_function, TYPE_PARAMS, ())

    assert not is_generic_function(generic_function)

    setattr(generic_function, TYPE_PARAMS, (TypeVar("T"),))

    assert is_generic_function(generic_function)


def test_get_executing_function():

    def test_function(arg1: int, arg2: int) -> tuple[AnyFunction, TypeArgs] | None:
        if fn := get_executing_function():
            args = get_generic_parameters(fn)
            return fn, args


    assert not is_generic_function(test_function)

    result = test_function(1,2)

    assert result

    fn, args = result

    assert fn is test_function
    assert not args # this is not a generic function (which is not available until python 3.12)


    def generic_function(arg1: int, arg2: int) -> tuple[AnyFunction, TypeArgs] | None:
        if fn := get_executing_function():
            args = get_generic_parameters(fn)
            return fn, args

    setattr(generic_function, TYPE_PARAMS, (TypeVar("T"),))

    result = generic_function(1,2)

    assert result

    fn, args = result

    assert fn is generic_function
    assert args # this is not a generic function (which is not available until python 3.12)



INIT = "__init__"
NEW = "__new__"
CLASS = "__class__"
GENERIC_CONSTRUCTOR = "__class_getitem__"
SPECIAL_CONSTRUCTOR = "__getitem__"
ORIGIN = "__origin__"
ORIGINAL_CLASS = "__orig_class__"
ARGS = "__args__"
PARAMETERS = "__parameters__"
BASES = "__bases__"
NAME = "__name__"
QUALIFIED_NAME = "__qualname__"
MODULE = "__module__"
ANNOTATIONS = "__annotations__"
DICT = "__dict__"
SELF = "__self__"
GLOBALS = "__globals__"
IM_CLASS = "im_class"
WRAPPED = "__wrapped__"
BOUND = "__bound__"
CONSTRAINTS = "__constraints__"
TYPE_PARAMS = "__type_params__"
DEFAULT = "__default__"


from typingutils.core.instances import get_original_class
from typingutils.core.functions import get_executing_function
from typingutils.core.types import (
    get_generic_origin, get_types_from_typevar, get_union_types,
    construct_generic_type
)

__all__ = [
    'get_generic_origin',
    'get_union_types',
    'get_original_class',
    'get_executing_function',
    'get_types_from_typevar',
    'construct_generic_type',
]


# pyright: basic
from typing import Any, Generic, TypeVar, Union, Optional, List, Dict, cast
from types import NoneType, EllipsisType, UnionType
from pytest import raises as assert_raises

from typingutils import (
    is_type, is_union, get_type_name, is_optional, get_optional_type, TypeParameter, UnionParameter, TypeVarParameter,
    is_subscripted_generic_type, is_generic_type, get_generic_arguments, get_generic_parameters
)
from typingutils.core.compat.typevar_tuple import TypeVarTuple
from typingutils.internal import get_generic_origin, get_original_class, get_union_types, get_types_from_typevar, construct_generic_type
from typingutils.core.attributes import ORIGIN

from tests.testcase_generators.issubclass import create_testcases_for_issubclass
from tests.testcase_generators.types import create_testcases_for_types
from tests.generic_types import non_generic_types, generic_types
from tests.generic_classes import ParentClass, GenericClass1


issubclass_testcases = list(create_testcases_for_issubclass())
types_testcases = list(create_testcases_for_types())

def test_is_type():
    tested: set[type[Any]] = set()

    assert not is_type(object)

    for testcase in types_testcases:

        if testcase.cls not in tested:
            tested.add(testcase.cls)

            result = is_type(testcase.cls)
            print(f"Testing is_type({get_type_name(testcase.cls)}) ==> {result}")
            if not result:
                is_type(testcase.cls)
            assert result


def test_get_generic_parameters():
    tested: set[type[Any]] = set()

    for testcase in types_testcases:
        if testcase.cls not in tested:
            tested.add(testcase.cls)

            result = get_generic_parameters(testcase.cls)

            print(f"Testing get_generic_parameters({get_type_name(testcase.cls)}) ==> {result}")

            if testcase.is_generic_type:
                assert any(result)
            else:
                assert not any(result)




def test_get_generic_arguments():
    tested: set[type[Any]] = set()

    for testcase in types_testcases:
        if testcase.cls not in tested:
            tested.add(testcase.cls)

            result = get_generic_arguments(testcase.cls)

            print(f"Testing get_generic_arguments({get_type_name(testcase.cls)}) ==> {result}")

            if testcase.is_generic: # non-generic objects like types may still have generic arguments (values in __args__ attribute)
                if not any(result):
                    get_generic_arguments(testcase.cls)
                assert any(result)





def test_get_generic_origin():
    tested: set[type[Any]] = set()

    for testcase in types_testcases:
        if testcase.cls not in tested:
            tested.add(testcase.cls)

            result = get_generic_origin(testcase.cls)

            if testcase.is_generic:
                origin = getattr(testcase.cls, ORIGIN)
                print(f"Testing get_generic_origin({get_type_name(testcase.cls)}) ==> {get_type_name(result)}")
                assert result == origin

            elif testcase.is_generic_type:
                if hasattr(testcase.cls, ORIGIN): # python 3.10 always has the __orig__ attribute
                    origin = getattr(testcase.cls, ORIGIN)
                    print(f"Testing get_generic_origin({get_type_name(testcase.cls)}) ==> {get_type_name(result)}")
                    assert result == origin
            else:
                result = get_generic_origin(testcase.cls)
                print(f"Testing get_generic_origin({get_type_name(testcase.cls)}) ==> {get_type_name(result)}")
                assert result == testcase.cls


    for cls, expected in (
        (str | int, UnionType),
    ):
        result = get_generic_origin(cls)
        print(f"Testing get_generic_origin({get_generic_origin(cls)}) ==> {result}")
        assert result == expected

    with assert_raises(ValueError):
        get_generic_origin(TypeVar("T"))



def test_is_generic_type():
    tested_base: set[type[Any]] = set()
    tested_comparison: set[type[Any]] = set()

    assert not is_generic_type(TypeVar("T"))
    assert is_generic_type(GenericClass1)

    for testcase in issubclass_testcases:
        if testcase.base not in tested_base:
            tested_base.add(testcase.base)

            if testcase.base_is_generic_type:
                result = is_generic_type(testcase.base)
                print(f"Testing is_generic_type({get_type_name(testcase.base)}) ==> {result}")
                assert result
            else:
                result = is_generic_type(testcase.base)
                print(f"Testing !is_generic_type({get_type_name(testcase.base)}) ==> {result}")
                assert not result


        if testcase.comparison not in tested_comparison:
            tested_comparison.add(testcase.comparison)

            if testcase.comparison_is_generic_type:
                result = is_generic_type(testcase.comparison)
                print(f"Testing is_generic_type({get_type_name(testcase.comparison)}) ==> {result}")
                assert result
            else:
                result = is_generic_type(testcase.comparison)
                print(f"Testing !is_generic_type({get_type_name(testcase.comparison)}) ==> {result}")
                assert not result


def test_is_subscripted_generic_type():
    tested: set[type[Any]] = set()

    for cls in non_generic_types:
        assert not is_subscripted_generic_type(cls)

    for testcase in issubclass_testcases:
        if testcase.base not in tested:
            tested.add(testcase.base)

            result = is_subscripted_generic_type(testcase.base)

            if testcase.base_is_subscripted_generic:
                print(f"Testing is_subscripted_generic_type({get_type_name(testcase.base)}) ==> {result}")
                assert result
            else:
                print(f"Testing !is_subscripted_generic_type({get_type_name(testcase.base)}) ==> {result}")
                assert not result


        if testcase.comparison not in tested:
            tested.add(testcase.comparison)

            result = is_subscripted_generic_type(testcase.comparison)

            if testcase.base_is_subscripted_generic:
                print(f"Testing is_generic({get_type_name(testcase.comparison)}) ==> {result}")
                assert result
            else:
                print(f"Testing !is_generic({get_type_name(testcase.comparison)}) ==> {result}")
                assert not result



def test_get_type_name():
    tested: set[type[Any] | TypeVarParameter ] = set()

    for cls in set([
        *generic_types,
        *non_generic_types,
        *( t.base for t in issubclass_testcases ),
        *( t.comparison for t in issubclass_testcases ),
        TypeVarTuple("Ts")
    ]):
        if cls not in tested:
            tested.add(cls)
            result = get_type_name(cls)

            print(f"Testing get_type_name({get_type_name(cls)}) ==> {result}")

            assert len(result) > 0
            assert not result.startswith(".") or result == "..."
            assert not result.startswith("<class")

    for cls, expected in cast(tuple[tuple[TypeParameter|UnionParameter, str]], (
        (EllipsisType, "..."),
        (NoneType, "None"),
        (List[str], "typing.List[str]"),
        (list[str], "list[str]"),
        (Dict[str, int], "typing.Dict[str, int]"),
        (dict[str, int], "dict[str, int]"),
        ((str, int), "(str, int)"),
        (str | int, "str | int"),
        (str | int | None, "str | int | None"),
        (ParentClass, "tests.generic_classes.ParentClass"),
        (ParentClass.ChildClass, "tests.generic_classes.ParentClass.ChildClass"),
    )):
        result = get_type_name(cls)
        print(f"Testing get_type_name({get_type_name(cls)}) ==> {result}")
        assert result == expected

    for cls in (
        TypeVar("T"),
        TypeVar("T2", bound=str),
        TypeVar("T3", str, int)
    ):

        result = get_type_name(cls)
        print(f"Testing get_type_name({get_type_name(cls)}) ==> {result}")
        assert result.startswith("~T")


def test_is_union():
    for cls, expected in cast(tuple[tuple[TypeParameter|UnionParameter, bool]], (
        (Optional[int], True),
        (Optional[Union[int, str]], True),
        (Union[int, None], True),
        (Union[int, str], True),
        (Union[int, str, None], True),
        (str, False),
        (int|str, True),
        (int|str|None, True),
    )):
        result = is_union(cls)
        print(f"Testing is_union({get_type_name(cls)}) ==> {result}")
        assert result == expected


def test_get_union_types():
    for cls, expected in cast(tuple[tuple[TypeParameter|UnionParameter, tuple[TypeParameter, ...]]], (
        (Union[int, None], (int,)),
        (Union[int, str], (int, str)),
        (Union[int, str, None], (int, str)),
        (int|None, (int,)),
        (int|str, (int, str)),
        (int|str|None, (int, str)),
    )):
        result = get_union_types(cast(UnionParameter, cls))
        print(f"Testing get_union_types({get_type_name(cls)}) ==> {result}")
        assert result == expected


def test_is_optional():
    for cls, expected in cast(tuple[tuple[TypeParameter|UnionParameter, bool]], (
        (str, False),
        (Optional[str], True),
        (Union[str, None], True),
        (Union[str, int], False),
        (Union[str, int, None], True),
        (str|None, True),
        (str|int, False),
        (str|int|None, True),
    )):
        result = is_optional(cls)
        print(f"Testing is_optional({get_type_name(cls)}) ==> {result}")
        assert result == expected


def test_get_optional_type():
    for cls, expected in cast(tuple[tuple[TypeParameter|UnionParameter, tuple[TypeParameter|UnionParameter, bool]]], (
        (Union[str, list[str]], (Union[str, list[str]], False)),
        (Optional[list[str]], (list[str], True)),
        (Union[str, None], (str, True)),
        (Union[int, str, None], (Union[int, str], True)),
        (Optional[str], (str, True)),
        (str, (str, False)),
        (str|list[str], (str|list[str], False)),
        (str|None, (str, True)),
        (int|str|None, (int|str, True)),
    )):
        result = get_optional_type(cls)
        print(f"Testing get_optional_type({get_type_name(cls)}) ==> {result}")
        assert result == expected

def test_get_types_from_typevar():
    for typevar, expected in cast(tuple[tuple[TypeVar, TypeParameter]], (
        (TypeVar("T1"), type[Any]),
        (TypeVar("T1", bound=str), str),
        (TypeVar("T1", str, int), str|int),
        (TypeVar("T1", bound=int), int),
        (TypeVar("T1", bool, int), bool|int),
        (TypeVarTuple("Tv1"), tuple[type[Any], ...]),
    )):
        result = get_types_from_typevar(typevar)
        print(f"Testing get_types_from_typevar({get_type_name(typevar)}) ==> {result}")
        assert result == expected

def test_construct_generic_type():
    for cls, types, expected in cast(tuple[tuple[TypeParameter, tuple[TypeParameter, ...], TypeParameter]], (
        (list, (str,), list[str]),
        (List, (str,), List[str]),
        (Optional, (str,), Optional[str]),
        (tuple, (str,), tuple[str]),
        (tuple, (str, ...), tuple[str, ...]),
        (dict, (str, int), dict[str, int]),
        (Dict, (str, int), Dict[str, int]),
    )):

        result = construct_generic_type(cls, *types)
        print(f"Testing construct_generic_type({get_type_name(cls)}, {', '.join((get_type_name(t) for t in types))}) ==> {result}")
        assert result == expected



def test_get_original_class():
    T = TypeVar("T")
    class GenericClass(Generic[T]): ...
    gc = GenericClass[str]()

    assert get_original_class(gc) == GenericClass[str]




from inspect import stack as get_stack

from typingutils.core.attributes import TYPE_PARAMS
from typingutils.core.types import AnyFunction



def is_generic_function(fn: AnyFunction) -> bool:
    """
    Python 3.12 or newer.

    Indicates whether or not type is a generic function.

    Args:
        fn (AnyFunction): A function.

    Returns:
        bool: A boolean indicating if function is generic or not.
    """

    if hasattr(fn, TYPE_PARAMS):
        from typingutils.core.instances import _extract_args # pyright: ignore[reportPrivateUsage]
        params, *_ = _extract_args(fn)
        if params is not None and any(params):
            return True
    return False

def get_executing_function() -> AnyFunction | None:
    """
    Returns the executing function from within it, enabling retrieval of its generic parameters.

    Returns:
        AnyFunction | None: The function executing
    """
    stack = get_stack()
    name: str | None = None

    for frame in stack[1:]:
        if frame.filename == __file__:
            continue # pragma: no cover

        locals = frame[0].f_locals

        if not name:
            name = frame.function
        elif name in locals:
            return locals[name]

        else:
            break # pragma: no cover

    return None # pragma: no cover



# pyright: basic
from typing import Any, Type, Iterable, TypeVar, Generic, cast
from collections import abc, deque, defaultdict, OrderedDict, ChainMap
from types import NoneType
from pytest import raises as assert_raises, fixture
from enum import Enum

from typingutils import issubclass_typing, get_type_name, get_type_name, TypeParameter, UnionParameter

from tests.other_impl.issubclass import comparison_generator
from tests.testcase_generators.issubclass import create_testcases_for_issubclass
from tests.generic_classes import DerivedClass1, DerivedClass2, ParentClass, BaseClass, Enumeration


subclass_assertions: list[tuple[TypeParameter | tuple[TypeParameter, ...], tuple[TypeParameter | UnionParameter, ...]]] = [
    (type, (
        object,
        Type,
        Type[Any],
        type,
        type[Any])),
    (NoneType, (
        object,
        NoneType,
        Type[Any],
        type[Any])),
    (list, (
        object,
        abc.Collection)),
    (deque, (
        object,
        Type[Any],
        type[Any],
        deque,
        abc.Collection)),
    (dict, (
        object,
        Type[Any],
        type[Any],
        dict,
        abc.Mapping,
        abc.Collection)),
    (defaultdict, (
        object,
        Type[Any],
        type[Any],
        dict,
        defaultdict,
        abc.Mapping,
        abc.MutableMapping,
        abc.Collection)),
    (OrderedDict, (
        object,
        Type[Any],
        type[Any],
        dict,
        OrderedDict,
        abc.Mapping,
        abc.MutableMapping,
        abc.Collection)),
    (ChainMap, (
        object,
        Type[Any],
        type[Any],
        ChainMap,
        abc.Mapping,
        abc.MutableMapping,
        abc.Collection)),
    (set, (
        object,
        Type[Any],
        type[Any],
        set,
        abc.Set,
        abc.Collection)),
    (frozenset, (
        object,
        Type[Any],
        type[Any],
        abc.Set,
        abc.Collection)),
    (tuple, (
        object,
        Type[Any],
        type[Any],
        tuple)),
    (tuple[str], (
        object,
        Type[Any],
        type[Any],
        tuple[Any],
        tuple)),
    ((str, int, float, bool), (
        object,
        Type[Any],
        type[Any],
        (str, int, float, bool)
    )),
    (DerivedClass2, (
        ParentClass,
        BaseClass
    )),
    (Enumeration, (
        object,
        Enum
    ))
]

all_types = set([ type_ for _, types in subclass_assertions for type_ in types ])

@fixture(scope = "module")
def comparisons():
    impl: dict[str, list[tuple[str, str]]] = defaultdict(lambda: [])
    yield impl

    print("\n")

    for key in impl:
        count = 0
        failed = 0
        for comparison, result in impl[key]:
            if "Error" in result:
                failed += 1
                print(f"{comparison} ==> FAILED")
            else:
                print(f"{comparison} ==> {result}")
                count +=1

        print(f"Comparison with {key}: {count} differences and {failed} failed\n")


issubclass_testcases = list(create_testcases_for_issubclass())

def test_multiple_bases(comparisons: dict[str, list[tuple[str, str]]]):
    assert issubclass_typing(DerivedClass2, ParentClass)
    assert issubclass_typing(DerivedClass2, BaseClass)
    assert not issubclass_typing(DerivedClass2, DerivedClass1)

def test_issubclass_typing(comparisons: dict[str, list[tuple[str, str]]]):
    tested_base: set[type[Any]] = set()
    tested_comparison: set[type[Any]] = set()

    for testcase in issubclass_testcases:
        if testcase.base not in tested_base:
            tested_base.add(testcase.base)
            result = issubclass_typing(testcase.base, testcase.base)
            print(f"Testing issubclass_typing({get_type_name(testcase.base)}, {get_type_name(testcase.base)}) ==> {result}")
            assert result

            for impl, result_comparison in comparison_generator(testcase.base, testcase.base):
                if result != result_comparison:
                    comparisons[impl].append((f"Comparing {impl}.issubclass({get_type_name(testcase.base)}, {get_type_name(testcase.base)})", f"{result_comparison} != {result}"))

        if testcase.comparison not in tested_comparison:
            tested_comparison.add(testcase.comparison)

            result = issubclass_typing(testcase.comparison, testcase.base)

            if testcase.expected_equality:
                print(f"Testing issubclass_typing({get_type_name(testcase.comparison)}, {get_type_name(testcase.base)}) ==> {result}")
                assert result
            else:
                print(f"Testing !issubclass_typing({get_type_name(testcase.comparison)}, {get_type_name(testcase.base)}) ==> {result}")
                assert not result

            for impl, result_comparison in comparison_generator(testcase.comparison, testcase.base):
                if result != result_comparison:
                    comparisons[impl].append((f"Comparing {impl}.issubclass({get_type_name(testcase.comparison)}, {get_type_name(testcase.base)})", f"{result_comparison} != {result}"))


def test_explicit_assertions(comparisons: dict[str, list[tuple[str, str]]]):
    for test_types, types in subclass_assertions:
        for test_type in cast(Iterable[type], test_types if isinstance(test_types, tuple) else (test_types,)):
            for type_ in types:
                result = issubclass_typing(test_type, type_)


                print(f"Testing issubclass_typing({get_type_name(test_type)}, {get_type_name(type_)}) ==> {result}")
                if not result:
                    issubclass_typing(test_type, type_)
                assert result

                for impl, result_comparison in comparison_generator(test_type, type_):
                    if result != result_comparison:
                        comparisons[impl].append((f"Comparing {impl}.issubclass({get_type_name(test_type)}, {get_type_name(type_)})", f"{result_comparison} != {result}"))


def test_tuple_bases(comparisons: dict[str, list[tuple[str, str]]]):
    for cls, base, expected in cast(tuple[tuple[TypeParameter, TypeParameter|tuple[TypeParameter], bool]], (
        (str, (str, int, bool), True),
        (str, (int, bool), False),
    )):
        result = issubclass_typing(cls, base)
        print(f"Testing issubclass_typing({get_type_name(cls)}, {get_type_name(base)}) ==> {result}")
        assert result == expected

        for impl, result_comparison in comparison_generator(cls, base):
            if result != result_comparison:
                comparisons[impl].append((f"Comparing {impl}.issubclass({get_type_name(cls)}, {get_type_name(base)})", f"{result_comparison} != {result}"))


def test_union_bases(comparisons: dict[str, list[tuple[str, str]]]):
    for cls, base, expected in cast(tuple[tuple[TypeParameter, TypeParameter|tuple[TypeParameter], bool]], (
        (str, int | bool, False),
        (str, str | float, True),
    )):
        result = issubclass_typing(cls, base)
        print(f"Testing issubclass_typing({get_type_name(cls)}, {get_type_name(base)}) ==> {result}")
        assert result == expected

        for impl, result_comparison in comparison_generator(cls, base):
            if result != result_comparison:
                comparisons[impl].append((f"Comparing {impl}.issubclass({get_type_name(cls)}, {get_type_name(base)})", f"{result_comparison} != {result}"))


def test_typevars(comparisons: dict[str, list[tuple[str, str]]]):
    for cls, base, expected in (
        (dict[str, int], dict[TypeVar("T1"), TypeVar("T2")], True),
        (dict[str, int], dict[TypeVar("T1", bound=str), TypeVar("T2", bound=int)], True),
        (dict[str, int], dict[TypeVar("T1", bound=bool), TypeVar("T2", bound=int)], False),
        (str, TypeVar("T"), True),
        (str, TypeVar("T", int, str), True),
        (str, TypeVar("T", bound=str), True),
        (str, TypeVar("T", bound=int), False),
    ):

        result = issubclass_typing(cls, base)
        print(f"Testing issubclass_typing({get_type_name(cls)}, {get_type_name(base)}) ==> {result}")

        assert result == expected

        for impl, result_comparison in comparison_generator(cls, base):
            if result != result_comparison:
                comparisons[impl].append((f"Comparing {impl}.issubclass({get_type_name(cls)}, {get_type_name(base)})", f"{result_comparison} != {result}"))


def test_multilevel_types(comparisons: dict[str, list[tuple[str, str]]]):
    for cls, base, expected in (
        (tuple[list[str]], tuple[list[str]], True),
        (tuple[list[str]], tuple[list[int]], False),
        (tuple[list[dict[str, int]]], tuple[list[dict[str, int]]], True),
        (tuple[list[dict[str, int]]], tuple[list[dict[str, bool]]], False),
        (tuple[list[dict[str, set[int]]]], tuple[list[dict[str, set[int]]]], True),
        (tuple[list[dict[str, set[int]]]], tuple[list[dict[str, set[str]]]], False),
    ):

        result = issubclass_typing(cls, base)
        print(f"Testing issubclass_typing({get_type_name(cls)}, {get_type_name(base)}) ==> {result}")

        assert result == expected

        for impl, result_comparison in comparison_generator(cls, base):
            if result != result_comparison:
                comparisons[impl].append((f"Comparing {impl}.issubclass({get_type_name(cls)}, {get_type_name(base)})", f"{result_comparison} != {result}"))





from typing import Annotated, Type, TypeVar, Sequence, Any, Generic, Union, Callable, cast, overload
from typing import _GenericAlias, GenericAlias, _SpecialGenericAlias, _UnionGenericAlias, _SpecialForm # pyright: ignore[reportUnknownVariableType, reportAttributeAccessIssue, reportPrivateUsage ]
from types import UnionType, NoneType, EllipsisType, FunctionType
from collections import abc

from typingutils.core.compat.typevar_tuple import TypeVarTuple
from typingutils.core.attributes import (
    BASES, NAME, QUALIFIED_NAME, MODULE, ORIGIN, ARGS, PARAMETERS, TYPE_PARAMS,
    GENERIC_CONSTRUCTOR, SPECIAL_CONSTRUCTOR, BOUND, CONSTRAINTS
)

TypeParameter = Annotated[ type | type[Any], "Represents any type." ]
UnionParameter = Annotated[ UnionType | tuple[TypeParameter, ...], "Represents a union of types." ]
TypeVarParameter = Annotated[ TypeVar | TypeVarTuple, "Represents a generic type parameter." ]
AnyType = Annotated[ TypeParameter | UnionParameter | TypeVarParameter, "Represents any type, union or typevar." ]
TypeArgs = Annotated[ tuple[AnyType, ...], "Represents a sequence of types, unions or typevars." ]
AnyFunction = Annotated[ FunctionType | Callable[..., Any], "Represents any function" ]

SetOfAny = set((Any,))

GENERIC_BASE_TYPES: set[type] = set(cast(list[type], [ _GenericAlias, GenericAlias, _SpecialGenericAlias ]))
GENERIC_SPECIAL_TYPES: set[type] = set(cast(list[type], [ _SpecialGenericAlias, _SpecialForm, _UnionGenericAlias ]))


def is_generic_type(cls: AnyType) -> bool:
    """
    Indicates whether or not type is a generic type, i.e. a type capable of subscripted generics like list[T].

    Args:
        cls (AnyType): A type.

    Returns:
        bool: A boolean indicating if type is a generic type or not.
    """

    if type(cls) is TypeVar:
        return False
    elif hasattr(cls, ORIGIN):
        from typingutils.core.instances import _extract_args # pyright: ignore[reportPrivateUsage]
        parameters, args, _ = _extract_args(cls)
        return any(parameters) if parameters is not None else args is None
    elif hasattr(cls, BASES) and Generic in getattr(cls, BASES):
        return True
    elif type(cls) in GENERIC_BASE_TYPES:
        return True # pragma: no cover
    else:
        return False


def _is_special_generic_type(cls: AnyType) -> bool:
    """
    Indicates if type is a special generic type from the typing module.

    Args:
        cls (AnyType): A type.

    Returns:
        bool: A boolean indicating if type is a special generic type or not.
    """

    return type(cls) in GENERIC_SPECIAL_TYPES # pragma: no cover


def is_subscripted_generic_type(cls: AnyType) -> bool:
    from typingutils.core.instances import get_generic_arguments

    return hasattr(cls, ARGS) and any(get_generic_arguments(cls))

@overload
def get_generic_parameters(obj: TypeParameter | AnyFunction) -> tuple[TypeVar, ...]:
    ...
@overload
def get_generic_parameters(obj: TypeParameter | AnyFunction, *, extract_types_from_typevars: bool = False) -> TypeArgs:
    ...
def get_generic_parameters(obj: TypeParameter | AnyFunction, *, extract_types_from_typevars: bool = False) -> TypeArgs:
    """
    Returns the generic typevars required to create a subscripted generic type (or function, python 3.12).
    of an object. Will even work when called from within a constructor of the class.

    Examples:
        T = TypeVar('T')
        class GenClass(Generic[T]): pass
        a = get_generic_parameters(GenClass[str]) => ~T

    Args:
        obj (TypeParameter | AnyFunction): A type or function.

    Returns:
        TypeArgs: A sequence of types.
    """
    for attr in (PARAMETERS, TYPE_PARAMS): # __type_params__ is new in python 3.12
        if hasattr(obj, attr):
            if params := getattr(obj, attr):
                return tuple(
                    get_types_from_typevar(param) if extract_types_from_typevars and isinstance(param, TypeVar) else param
                    for param in params
                )

    return ()


def get_type_name(cls: AnyType) -> str:
    """
    Gets the name of a type.

    Args:
        cls (AnyType): A type.

    Raises:
        Exception: An exception is raised if a type name cannot be produced.

    Returns:
        str: The name of the type.
    """

    if cls is None or cls == NoneType: # pyright: ignore[reportUnnecessaryComparison]
        return "None"
    if cls is EllipsisType:
        return "..."
    if isinstance(cls, TypeVarTuple):
        return f"*{cls}"
    if isinstance(cls, TypeVar):
        if hasattr(cls, BOUND) and ( bound := getattr(cls, BOUND) ):
            return f"{cls}<{get_type_name(bound)}>"
        elif hasattr(cls, CONSTRAINTS) and ( constraints := tuple( get_type_name(t) for t in getattr(cls, CONSTRAINTS) ) ):
            return f"{cls}<{'|'.join(constraints)}>"
        else:
            return f"{cls}"
    if isinstance(cls, tuple):
        return f"({', '.join( get_type_name(item) for item in cls )})"


    from typingutils.core.instances import get_generic_arguments
    args: Sequence[str] = [
        get_type_name(arg)
        for arg in get_generic_arguments(cls)
    ]

    if is_union(cls):
        return ' | '.join(args)

    type_name: str | None = None

    name = cls.__qualname__ if hasattr(cls, QUALIFIED_NAME) else getattr(cls, NAME) if hasattr(cls, NAME) else str(cls)

    if hasattr(cls, MODULE) and getattr(cls, MODULE) in [ "builtins", "__builtins__" ]:
        type_name = name
    elif hasattr(cls, MODULE) and not name.startswith(getattr(cls, MODULE)):
        type_name =  f"{getattr(cls, MODULE)}.{name}"
    else:
        type_name = name

    if args:
        return f"{type_name}[{', '.join(args)}]"
    elif type_name:
        return type_name
    else:
        raise Exception(f"Unable to get name of type {cls}") # pragma: no cover


def get_generic_origin(cls: AnyType) -> TypeParameter:
    """
    Gets the generic origin of a type i.e. the type a generic type originates from.

    Args:
        cls (AnyType): A type.

    Raises:
        ValueError: _description_

    Returns:
        TypeParameter: A type.
    """
    if type(cls) is TypeVar:
        raise ValueError("TypeVar is not a generic type")
    elif hasattr(cls, ORIGIN):
        return cast(type, getattr(cls, ORIGIN))
    elif isinstance(cls, UnionType):
        return UnionType # pyright: ignore[reportReturnType]

    return cast(type, cls)


def get_union_types(cls: UnionParameter) -> TypeArgs:
    """
    Returns the types which make up a union.

    Args:
        cls (UnionParameter): A union of types.

    Raises:
        ValueError: A ValueError is raised if type is not a union.

    Returns:
        TypeArgs: A sequence of types.
    """

    if type(cls) is not TypeVar and hasattr(cls, ARGS): # pyright: ignore[reportUnnecessaryComparison]
        args = tuple([
            arg
            for arg in cast(tuple[Any], getattr(cls, ARGS))
            if arg != NoneType
        ])
        return args
    else:
        raise ValueError(f"Type {get_type_name(cls)} is not a union") # pragma: no cover


def is_union(cls: AnyType) -> bool:
    """
    Indicates whenter or not type is a union of types.

    Args:
        cls (AnyType): A type.

    Returns:
        bool: Returns true if type is a union of types.
    """

    if type(cls) is TypeVar:
        return False
    elif isinstance(cls, UnionType):
        return True
    else:
       return get_generic_origin(cast(TypeParameter, cls)) is Union


def is_optional(cls: AnyType) -> bool:
    """
    Indicates if type is an optional type, i.e. a union containing None or a typing.Optional[T].

    Args:
        cls (AnyType): A type.

    Returns:
        bool: True if type is optional.
    """

    if is_union(cls):
        classes = getattr(cls, ARGS)
        if NoneType in classes:
            return True
    elif isinstance(cls, TypeVar):
        return is_optional(get_types_from_typevar(cls))
    return False



def get_optional_type(cls: AnyType) -> tuple[AnyType, bool]:
    """
    Returns the actual type from an optional type.

    Examples:
        get_optional_type(str) # => (str, False)
        get_optional_type(Optional[str]) # => (str, True)
        get_optional_type(str | int | None) # => (str | int, True)

    Args:
        cls (AnyType): A type.

    Returns:
        tuple[AnyType, bool]: A tuple with the type and a boolean indicating if optional or not.
    """

    if is_union(cls):
        classes_org = getattr(cls, ARGS)
        optional: bool = NoneType in classes_org
        classes = [ c for c in classes_org if c != NoneType ]

        if len(classes) == 1:
            return classes[0], optional
        else:
            return cast(type, Union[tuple(classes)]), optional

    elif isinstance(cls, TypeVar):
        return get_optional_type(get_types_from_typevar(cls))

    return cast(type, cls), False


def issubclass_typing(cls: AnyType, base: AnyType | TypeArgs ) -> bool:
    """
    Indicates whether or not one type is a subclass of another. This implementation
    works similarly to the builtin issubclass(), but supports generics as well.

    Args:
        cls (AnyType): A type.
        base (AnyType | TypeArgs): A type or typevar.

    Raises:
        ValueError: A ValueError is raised if cls is a typevar.

    Returns:
        bool: True if cls is an instance of base.
    """

    if isinstance(cls, (TypeVar, TypeVarTuple)):
        raise ValueError("Argument cls cannot be an instance of TypeVar or TypeVarTuple")
    if not cls:
        return False # pragma: no cover
    elif cls is base:
        return True
    elif base in (object, TypeParameter,  type[Any], Type[Any]):
        return True

    from typingutils.core.instances import get_generic_arguments, check_type

    if isinstance(base, tuple):
        for base_cls in cast(Sequence[TypeParameter], base):
            if issubclass_typing(cls, base_cls):
                return True
        return False


    cls_is_type, cls_is_generic_type, cls_is_subscripted_generic_type = check_type(cls)
    base_is_type, base_is_generic_type, base_is_subscripted_generic_type = check_type(base)
    base_args: TypeArgs = ()

    if base_is_subscripted_generic_type:
        base_args = get_generic_arguments(base)

    if base_args and set(base_args) == SetOfAny:
        base = get_generic_origin(cast(TypeParameter | UnionParameter | TypeVar, base))

    if is_union(base):
        classes = get_generic_arguments(base)
        return any([ issubclass_typing(cls, b) for b in classes ]) or cls_is_type and get_generic_origin(cast(type, cls)) is Union

    if cls_is_type:
        if base_is_type:
            if get_generic_origin(cls) is base:
                return True

            cls_args = get_generic_arguments(cls) if cls_is_subscripted_generic_type else get_union_types(cast(UnionParameter, cls)) if is_union(cls) else ()
            if base_is_generic_type:
                base_args = get_generic_parameters(cast(type[Any], base), extract_types_from_typevars = True)

            if get_generic_origin(cls) == get_generic_origin(base):
                if cls_args == base_args:
                    return True
                elif cls_args and base_args and len(cls_args) == len(base_args) and not [
                    t for t in zip(cls_args, base_args) if not issubclass_typing(t[0], t[1])
                ]:
                    return True
                else:
                    return not any(base_args)
            elif cls_is_type and not cls_is_generic_type and not cls_is_subscripted_generic_type and not base_is_subscripted_generic_type and isinstance(base, type):
                return issubclass(cast(type[Any], cls), base)

            else:
                return False

        elif type(base) is TypeVar:
            return issubclass_typing(cls, get_types_from_typevar(base))

    if cls_is_generic_type or cls_is_subscripted_generic_type or base_is_generic_type or base_is_subscripted_generic_type: # pragma: no cover
        return False

    return issubclass(cls, base) # fallback # pyright: ignore[reportArgumentType] # pragma: no cover

def get_types_from_typevar(typevar: TypeVarParameter) -> TypeParameter | UnionParameter:
    if isinstance(typevar, TypeVar) and hasattr(typevar, BOUND):
        if bound := getattr(typevar, BOUND):
            return Union[bound]
        elif constraints := getattr(typevar, CONSTRAINTS):
            return Union[constraints]
    elif isinstance(typevar, (TypeVar, TypeVarTuple)): # pyright: ignore[reportUnnecessaryIsInstance]
        return tuple[type[Any], ...]
    return type[Any]

def construct_generic_type(cls: type, *generic_arguments: AnyType | EllipsisType) -> type:
    generic_arguments = tuple( arg if arg != EllipsisType else Ellipsis for arg in generic_arguments )

    if hasattr(cls, GENERIC_CONSTRUCTOR): # types derived from Generic[T]
        return cast(Callable[[tuple[type, ...]], type], getattr(cls, GENERIC_CONSTRUCTOR))(generic_arguments) # pyright: ignore[reportArgumentType]
    elif _is_special_generic_type(cls): # types like Optional[T] and Union[T, ...]
        if len(generic_arguments) == 1:
            return cast(Callable[[type], type], getattr(cls, SPECIAL_CONSTRUCTOR))(generic_arguments[0]) # pyright: ignore[reportArgumentType]
        else:
            return cast(Callable[[tuple[type, ...]], type], getattr(cls, SPECIAL_CONSTRUCTOR))(generic_arguments) # pyright: ignore[reportArgumentType]
    else:
        return cast(Callable[[tuple[type, ...]], type], getattr(cls, SPECIAL_CONSTRUCTOR))(generic_arguments) # pyright: ignore[reportArgumentType] # pragma: no cover




from typingutils.core.instances import (
    isinstance_typing, is_type, is_subscripted_generic_type,
    get_generic_arguments, check_type, TypeCheck
)
from typingutils.core.types import (
    get_type_name, issubclass_typing, is_optional, get_optional_type,
    is_generic_type, is_union, get_generic_parameters,
    TypeParameter, UnionParameter, AnyType, AnyFunction,
    TypeArgs, TypeVarParameter
)
from typingutils.core.functions import is_generic_function

__all__ = [
    'TypeParameter',
    'UnionParameter',
    'AnyType',
    'AnyFunction',
    'TypeArgs',
    'TypeVarParameter',

    'TypeCheck',

    'isinstance_typing',
    'issubclass_typing',
    'check_type',
    'is_type',
    'is_subscripted_generic_type',
    'is_generic_type',
    'is_generic_function',
    'is_optional',
    'is_union',
    'get_type_name',
    'get_optional_type',
    'get_generic_arguments',
    'get_generic_parameters',
]


# pyright: basic
import sys
from typing import Any

if sys.version_info >= (3, 11):

    from typing import TypeVarTuple # pyright: ignore[reportAssignmentType, reportAttributeAccessIssue]

else:

    from typing_extensions import TypeVarTuple # pyright: ignore[reportMissingModuleSource] # pragma: no cover




# pyright: basic
from typing import Generic, Tuple, Dict, Optional, Union, Any, cast
from datetime import date, time, datetime
from pytest import raises as assert_raises

from typingutils import isinstance_typing, issubclass_typing, get_generic_arguments, get_generic_parameters
from typingutils.internal import get_original_class

from tests.generic_classes import NormalClass, GenericClass1, T, T1, T2

def test_get_original_class():
    class GenericClass(Generic[T]):
        def __new__(cls, *args: Any, **kwargs: Any):
            cls.clsorig = get_original_class(cls)
            return object.__new__(cls, *args, **kwargs)
        def __init__(self):
            self.orig = get_original_class(self)

    for inst, override_fn, expected, test_negation in cast(tuple[tuple[object, bool, type[Any], bool], ...], (
        (GenericClass[str]().clsorig, False, GenericClass[str],True),
        (GenericClass[str]().orig, False, GenericClass[str],True),
        (GenericClass[str](), True, GenericClass[str],True),
        (list[str](), True, list, False), # builtin generic types aren't supported
    )):
        if test_negation:
            assert type(inst) != expected
        result = get_original_class(inst) if override_fn else inst
        assert result == expected


def test_get_generic_arguments():
    class GenericClass(Generic[T]):
        def __new__(cls, *args: Any, **kwargs: Any):
            cls.clsargs = get_generic_arguments(cls)
            return object.__new__(cls, *args, **kwargs)

        def __init__(self):
            self.args = get_generic_arguments(self)
            x=0

        @property
        def test(self) -> str:
            args = get_generic_arguments(self)
            return str(args)

    for inst, override_fn, expected in cast(tuple[tuple[object, bool, tuple[Any, ...]], ...], (
        (GenericClass[str], True, (str,)),
        (GenericClass[str](), True, (str,)),
        (GenericClass[str]().args, False, (str,)),
        (GenericClass[str]().clsargs, False, (str,)),
        (str, True, ()),
        ("", True, ()),
    )):
        result = get_generic_arguments(inst) if override_fn else inst
        assert result == expected



def test_isinstance_typing():
    classes_simple: dict[type[Any], list[Any]] = {
        str: [ "test", "" ],
        bool: [ True, False ],
        int: [ 1, 2, 3 ],
        float: [ 1.1, 1.2, 1.3 ],
        date: [ datetime.today() ],
        time: [ datetime.now().time() ],
        datetime: [ datetime.now() ]
    }

    for cls, examples in classes_simple.items():
        for example in examples:
            assert isinstance_typing(example, cls)

    assert isinstance_typing("dsfhfdg", (str, int))
    assert isinstance_typing(NormalClass, type)
    assert not isinstance_typing(GenericClass1[int](), GenericClass1[str])
    assert isinstance_typing(GenericClass1[str](), GenericClass1[str])
    assert not isinstance_typing(T, str)
    assert isinstance_typing("abc", Union[int, str])
    assert not isinstance_typing("abc", Union[int, float])
    assert not isinstance_typing(None, Union[int, str])
    assert isinstance_typing("abc", Union[str, None])
    assert isinstance_typing("abc", Optional[str])
    assert isinstance_typing(None, Union[str, None])
    assert isinstance_typing(None, Optional[str])
    assert not isinstance_typing(list[str], type)
    assert not isinstance_typing(list[str], str)
    assert not isinstance_typing(list[str], list[str])
    assert isinstance_typing("abc", int|str)
    assert not isinstance_typing("abc", int|float)
    assert not isinstance_typing(None, int|str)
    assert isinstance_typing("abc", str|None)
    assert isinstance_typing(None, str|None)


def test_issubclass_typing():

    assert not issubclass_typing(Tuple, Tuple[str, ...])
    assert issubclass_typing(Dict[str, int], Dict[T1, T2]) # pyright: ignore[reportGeneralTypeIssues]

    with assert_raises(ValueError):
        assert not issubclass_typing(T, T) # pyright: ignore[reportCallIssue, reportArgumentType]
    with assert_raises(ValueError):
        assert not issubclass_typing(T1, T2) # pyright: ignore[reportCallIssue, reportArgumentType]
    with assert_raises(ValueError):
        assert not issubclass_typing(T, str) # pyright: ignore[reportCallIssue, reportArgumentType]



# pyright: basic
import sys

if sys.version_info >= (3, 11):

    from typing import Unpack # pyright: ignore[reportAssignmentType, reportAttributeAccessIssue]

else:

    from typing_extensions import Unpack # pyright: ignore[reportMissingModuleSource] # pragma: no cover
